@@ -4,7 +4,7 @@
  * Copyright (c) 1999-2006 Ethan Galstad (nagios@nagios.org)
  * License: GPL
  *
- * Last Modified: 01-23-2006
+ * Last Modified: 04-06-2006
  *
  * Command line: nrpe -c <config_file> [--inetd | --daemon]
  *
@@ -52,16 +52,25 @@ int     server_port=DEFAULT_SERVER_PORT;
 char    server_address[16]="0.0.0.0";
 int     socket_timeout=DEFAULT_SOCKET_TIMEOUT;
 int     command_timeout=DEFAULT_COMMAND_TIMEOUT;
+char    *command_prefix=NULL;
 
 command *command_list=NULL;
 
 char    *nrpe_user=NULL;
 char    *nrpe_group=NULL;
 
+char    *allowed_hosts=NULL;
+
 char    *pid_file=NULL;
+int     wrote_pid_file=FALSE;
 
 int     allow_arguments=FALSE;
 
+int     allow_weak_random_seed=FALSE;
+
+int     sigrestart=FALSE;
+int     sigshutdown=FALSE;
+
 int     show_help=FALSE;
 int     show_license=FALSE;
 int     show_version=FALSE;
@@ -77,6 +86,8 @@ int main(int argc, char **argv){
 	char buffer[MAX_INPUT_BUFFER];
 #ifdef HAVE_SSL
 	DH *dh;
+	char seedfile[FILENAME_MAX];
+	int i,c;
 #endif
 
 	result=process_arguments(argc,argv);
@@ -164,22 +175,13 @@ int main(int argc, char **argv){
 		config_file[sizeof(config_file)-1]='\x0';
 	        }
 
-	/* read the config file */
-	result=read_config_file(config_file);	
-
-	/* exit if there are errors... */
-	if(result==ERROR){
-		syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
-		return STATE_CRITICAL;
-		}
+        /* generate the CRC 32 table */
+        generate_crc32_table();
 
 	/* initialize macros */
 	for(x=0;x<MAX_COMMAND_ARGUMENTS;x++)
 		macro_argv[x]=NULL;
 
-        /* generate the CRC 32 table */
-        generate_crc32_table();
-
 #ifdef HAVE_SSL
 	/* initialize SSL */
 	if(use_ssl==TRUE){
@@ -187,6 +189,26 @@ int main(int argc, char **argv){
 		SSLeay_add_ssl_algorithms();
 		meth=SSLv23_server_method();
 		SSL_load_error_strings();
+
+		/* use week random seed if necessary */
+		if(allow_weak_random_seed && (RAND_status()==0)){
+
+			if(RAND_file_name(seedfile,sizeof(seedfile)-1))
+				if(RAND_load_file(seedfile,-1))
+					RAND_write_file(seedfile);
+
+			if(RAND_status()==0){
+				syslog(LOG_ERR,"Warning: SSL/TLS uses a weak random seed which is highly discouraged");
+				srand(time(NULL));
+				for(i=0;i<500 && RAND_status()==0;i++){
+					for(c=0;c<sizeof(seedfile);c+=sizeof(int)){
+						*((int *)(seedfile+c))=rand();
+					        }
+					RAND_seed(seedfile,sizeof(seedfile));
+					}
+				}
+			}
+
 		if((ctx=SSL_CTX_new(meth))==NULL){
 			syslog(LOG_ERR,"Error: could not create SSL context.\n");
 			exit(STATE_CRITICAL);
@@ -213,6 +235,15 @@ int main(int argc, char **argv){
 	/* if we're running under inetd... */
 	if(use_inetd==TRUE){
 
+		/* read the config file */
+		result=read_config_file(config_file);	
+
+		/* exit if there are errors... */
+		if(result==ERROR){
+			syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
+			return STATE_CRITICAL;
+		        }
+
 		/* make sure we're not root */
 		check_privileges();
 
@@ -230,16 +261,6 @@ int main(int argc, char **argv){
 		/* we're a daemon - set up a new process group */
 		setsid();
 
-		/* ignore SIGHUP */
-		signal(SIGHUP, SIG_IGN);
-
-		/* write pid file */
-		if(write_pid_file()==ERROR)
-			return STATE_CRITICAL;
-
-		chdir("/");
-		/*umask(0);*/
-
 		/* close standard file descriptors */
 		close(0);
 		close(1);
@@ -250,14 +271,66 @@ int main(int argc, char **argv){
 		open("/dev/null",O_WRONLY);
 		open("/dev/null",O_WRONLY);
 
+		chdir("/");
+		/*umask(0);*/
+
+		/* handle signals */
+		signal(SIGQUIT,sighandler);
+		signal(SIGTERM,sighandler);
+		signal(SIGHUP,sighandler);
+
+		/* log info to syslog facility */
+		syslog(LOG_NOTICE,"Starting up daemon");
+
+		/* read the config file */
+		result=read_config_file(config_file);	
+
+		/* exit if there are errors... */
+		if(result==ERROR){
+			syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
+			return STATE_CRITICAL;
+		        }
+
+		/* write pid file */
+		if(write_pid_file()==ERROR)
+			return STATE_CRITICAL;
+		
 		/* drop privileges */
 		drop_privileges(nrpe_user,nrpe_group);
 
 		/* make sure we're not root */
 		check_privileges();
 
-		/* wait for connections */
-		wait_for_connections();
+		do{
+
+			/* reset flags */
+			sigrestart=FALSE;
+			sigshutdown=FALSE;
+
+			/* wait for connections */
+			wait_for_connections();
+
+			/* free all memory we allocated */
+			free_memory();
+
+			if(sigrestart==TRUE){
+
+				/* read the config file */
+				result=read_config_file(config_file);	
+
+				/* exit if there are errors... */
+				if(result==ERROR){
+					syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
+					return STATE_CRITICAL;
+				        }
+			        }
+	
+		        }while(sigrestart==TRUE && sigshutdown==FALSE);
+
+		/* remove pid file */
+		remove_pid_file();
+
+		syslog(LOG_NOTICE,"Daemon shutdown\n");
 	        }
 
 #ifdef HAVE_SSL
@@ -366,12 +439,17 @@ int read_config_file(char *filename){
 				return ERROR;
 			        }
 		        }
+		else if(!strcmp(varname,"command_prefix"))
+			command_prefix=strdup(varvalue);
 
 		else if(!strcmp(varname,"server_address")){
                         strncpy(server_address,varvalue,sizeof(server_address) - 1);
                         server_address[sizeof(server_address)-1]='\0';
                         }
 
+                else if(!strcmp(varname,"allowed_hosts"))
+			allowed_hosts=strdup(varvalue);
+
 		else if(strstr(input_line,"command[")){
 			temp_buffer=strtok(varname,"[");
 			temp_buffer=strtok(NULL,"]");
@@ -407,6 +485,9 @@ int read_config_file(char *filename){
 			        }
 		        }
 
+		else if(!strcmp(varname,"allow_weak_random_seed"))
+			allow_weak_random_seed=(atoi(varvalue)==1)?TRUE:FALSE;
+
 		else if(!strcmp(varname,"pid_file"))
 			pid_file=strdup(varvalue);
 
@@ -430,6 +511,7 @@ int read_config_dir(char *dirname){
 	char config_file[MAX_FILENAME_LENGTH];
 	DIR *dirp;
 	struct dirent *dirfile;
+	struct stat buf;
 	int result=OK;
 	int x;
 
@@ -443,19 +525,18 @@ int read_config_dir(char *dirname){
 	/* process all files in the directory... */
 	while((dirfile=readdir(dirp))!=NULL){
 
+		/* create the full path to the config file or subdirectory */
+		snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
+		config_file[sizeof(config_file)-1]='\x0';
+		stat(config_file, &buf);
+
 		/* process this if it's a config file... */
 		x=strlen(dirfile->d_name);
 		if(x>4 && !strcmp(dirfile->d_name+(x-4),".cfg")){
 
-#ifdef _DIRENT_HAVE_D_TYPE
 			/* only process normal files */
-			if(dirfile->d_type!=DT_REG)
+			if(!S_ISREG(buf.st_mode))
 				continue;
-#endif
-
-			/* create the full path to the config file */
-			snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
-			config_file[sizeof(config_file)-1]='\x0';
 
 			/* process the config file */
 			result=read_config_file(config_file);
@@ -465,18 +546,13 @@ int read_config_dir(char *dirname){
 				break;
 		        }
 
-#ifdef _DIRENT_HAVE_D_TYPE
 		/* recurse into subdirectories... */
-		if(dirfile->d_type==DT_DIR){
+		if(S_ISDIR(buf.st_mode)){
 
 			/* ignore current, parent and hidden directory entries */
 			if(dirfile->d_name[0]=='.')
 				continue;
 
-			/* create the full path to the config directory */
-			snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
-			config_file[sizeof(config_file)-1]='\x0';
-
 			/* process the config directory */
 			result=read_config_dir(config_file);
 
@@ -484,7 +560,6 @@ int read_config_dir(char *dirname){
 			if(result==ERROR)
 				break;
 		        }
-#endif
 		}
 
 	closedir(dirp);
@@ -550,7 +625,6 @@ void wait_for_connections(void){
 	int rc;
 	int sock, new_sd;
 	socklen_t addrlen;
-	char connecting_host[16];
 	pid_t pid;
 	int flag=1;
 	fd_set fdread;
@@ -569,6 +643,9 @@ void wait_for_connections(void){
 	        exit(STATE_CRITICAL);
 		}
 
+	/* socket should be non-blocking */
+	fcntl(sock,F_SETFL,O_NONBLOCK);
+
         /* set the reuse address flag so we don't get errors when restarting */
         flag=1;
         if(setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&flag,sizeof(flag))<0){
@@ -601,22 +678,16 @@ void wait_for_connections(void){
 	        exit(STATE_CRITICAL);
 		}
 
-	/* log info to syslog facility */
-        syslog(LOG_NOTICE,"Starting up daemon");
-
 	/* log warning about command arguments */
 #ifdef ENABLE_COMMAND_ARGUMENTS
 	if(allow_arguments==TRUE)
 		syslog(LOG_NOTICE,"Warning: Daemon is configured to accept command arguments from clients!");
 #endif
 
-	/* Trap signals */
-	signal(SIGQUIT,sighandler);
-	signal(SIGTERM,sighandler);
+	syslog(LOG_INFO,"Listening for connections on port %d\n",htons(myname.sin_port));
 
-	if(debug==TRUE){
-		syslog(LOG_DEBUG,"Listening for connections on port %d\n",htons(myname.sin_port));
-	        }
+	if(allowed_hosts)
+		syslog(LOG_INFO,"Allowing connections from: %s\n",allowed_hosts);
 
 	/* listen for connection requests - fork() if we get one */
 	while(1){
@@ -631,6 +702,10 @@ void wait_for_connections(void){
 			timeout.tv_usec=500000;
 			retval=select(sock+1,&fdread,NULL,&fdread,&timeout);
 
+			/* bail out if necessary */
+			if(sigrestart==TRUE || sigshutdown==TRUE)
+				break;
+
 			/* error */
 			if(retval<0)
 				continue;
@@ -645,6 +720,10 @@ void wait_for_connections(void){
 				if(errno==ENOBUFS)
 					continue;
 
+				/* bail out if necessary */
+				if(sigrestart==TRUE || sigshutdown==TRUE)
+					break;
+
 				/* retry */
 				if(errno==EWOULDBLOCK || errno==EINTR)
 					continue;
@@ -657,6 +736,10 @@ void wait_for_connections(void){
 			break;
 		        }
 
+		/* bail out if necessary */
+		if(sigrestart==TRUE || sigshutdown==TRUE)
+			break;
+
 		/* child process should handle the connection */
     		pid=fork();
     		if(pid==0){
@@ -677,6 +760,11 @@ void wait_for_connections(void){
 					return;
 				        }
 
+				/* handle signals */
+				signal(SIGQUIT,child_sighandler);
+				signal(SIGTERM,child_sighandler);
+				signal(SIGHUP,child_sighandler);
+
 				/* grandchild does not need to listen for connections, so close the socket */
 				close(sock);  
 
@@ -701,6 +789,31 @@ void wait_for_connections(void){
 				if(debug==TRUE)
 					syslog(LOG_DEBUG,"Connection from %s port %d",inet_ntoa(nptr->sin_addr),nptr->sin_port);
 
+                                /* is this is a blessed machine? */
+				if(allowed_hosts){
+
+					if(!is_an_allowed_host(inet_ntoa(nptr->sin_addr))){
+
+                                               /* log error to syslog facility */
+                                               syslog(LOG_ERR,"Host %s is not allowed to talk to us!",inet_ntoa(nptr->sin_addr));
+
+                                               /* log info to syslog facility */
+					       if(debug==TRUE)
+						       syslog(LOG_DEBUG,"Connection from %s closed.",inet_ntoa(nptr->sin_addr));
+
+					       /* close socket prior to exiting */
+                                               close(new_sd);
+
+					       exit(STATE_OK);
+				               }
+                                       else{
+
+                                               /* log info to syslog facility */
+                                               if(debug==TRUE)
+                                                       syslog(LOG_DEBUG,"Host address is in allowed_hosts");
+				               }
+				        }
+
 #ifdef HAVE_LIBWRAP
 
 				/* Check whether or not connections are allowed from this host */
@@ -726,12 +839,12 @@ void wait_for_connections(void){
 
 				/* log info to syslog facility */
 				if(debug==TRUE)
-					syslog(LOG_DEBUG,"Connection from %s closed.",connecting_host);
+					syslog(LOG_DEBUG,"Connection from %s closed.",inet_ntoa(nptr->sin_addr));
 
 				/* close socket prior to exiting */
 				close(new_sd);
 
-				return;
+				exit(STATE_OK);
     			        }
 
 			/* first child returns immediately, grandchild is inherited by INIT process -> no zombies... */
@@ -749,11 +862,80 @@ void wait_for_connections(void){
 		        }
   		}
 
+	/* close the socket we're listening on */
+	close(sock);
+
 	return;
 	}
 
 
 
+/* checks to see if a given host is allowed to talk to us */
+int is_an_allowed_host(char *connecting_host){
+	char *temp_buffer=NULL;
+	char *temp_ptr=NULL;
+	int result=0;
+        struct hostent *myhost;
+	char **pptr=NULL;
+	char *save_connecting_host=NULL;
+	struct in_addr addr;
+	
+	/* make sure we have something */
+	if(connecting_host==NULL)
+		return 0;
+	if(allowed_hosts==NULL)
+		return 1;
+
+	if((temp_buffer=strdup(allowed_hosts))==NULL)
+		return 0;
+	
+	/* try and match IP addresses first */
+	for(temp_ptr=strtok(temp_buffer,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
+
+		if(!strcmp(connecting_host,temp_ptr)){
+			result=1;
+			break;
+		        }
+	        }
+
+	/* try DNS lookups if needed */
+	if(result==0){
+
+		free(temp_buffer);
+		if((temp_buffer=strdup(allowed_hosts))==NULL)
+			return 0;
+
+		save_connecting_host=strdup(connecting_host);
+		for(temp_ptr=strtok(temp_buffer,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
+
+			myhost=gethostbyname(temp_ptr);
+			if(myhost!=NULL){
+
+				/* check all addresses for the host... */
+				for(pptr=myhost->h_addr_list;*pptr!=NULL;pptr++){
+					memcpy(&addr, *pptr, sizeof(addr));
+					if(!strcmp(save_connecting_host, inet_ntoa(addr))){
+						result=1;
+						break;
+					        }
+				        }
+			        }
+
+			if(result==1)
+				break;
+		        }
+
+		strcpy(connecting_host, save_connecting_host);
+		free(save_connecting_host);
+	        }
+
+	free(temp_buffer);
+
+	return result;
+        }
+
+
+
 /* handles a client connection */
 void handle_connection(int sock){
         u_int32_t calculated_crc32;
@@ -791,7 +973,11 @@ void handle_connection(int sock){
 	if(result==STATE_OK && use_ssl==TRUE){
 		if((ssl=SSL_new(ctx))!=NULL){
 			SSL_set_fd(ssl,sock);
-			if((rc=SSL_accept(ssl))!=1){
+
+			/* keep attempting the request if needed */
+                        while(((rc=SSL_accept(ssl))!=1) && (SSL_get_error(ssl,rc)==SSL_ERROR_WANT_READ));
+
+			if(rc!=1){
 				syslog(LOG_ERR,"Error: Could not complete SSL handshake. %d\n",SSL_get_error(ssl,rc));
 #ifdef DEBUG
 				errfp=fopen("/tmp/err.log","w");
@@ -817,8 +1003,9 @@ void handle_connection(int sock){
 	if(use_ssl==FALSE)
 		rc=recvall(sock,(char *)&receive_packet,&bytes_to_recv,socket_timeout);
 #ifdef HAVE_SSL
-	else
-		rc=SSL_read(ssl,&receive_packet,bytes_to_recv);
+	else{
+                while(((rc=SSL_read(ssl,&receive_packet,bytes_to_recv))<=0) && (SSL_get_error(ssl,rc)==SSL_ERROR_WANT_READ));
+		}
 #endif
 
 	/* recv() error or client disconnect */
@@ -915,7 +1102,10 @@ void handle_connection(int sock){
 		else{
 
 			/* process command line */
-			strncpy(raw_command,temp_command->command_line,sizeof(raw_command)-1);
+			if(command_prefix==NULL)
+				strncpy(raw_command,temp_command->command_line,sizeof(raw_command)-1);
+			else
+				snprintf(raw_command,sizeof(raw_command)-1,"%s %s",command_prefix,temp_command->command_line);
 			raw_command[sizeof(raw_command)-1]='\x0';
 			process_macros(raw_command,processed_command,sizeof(processed_command));
 
@@ -1007,21 +1197,6 @@ void handle_connection(int sock){
 
 
 
-/* handle signals */
-void sighandler(int sig){
-
-	/* free all memory we allocated */
-	free_memory();
-	
-	/* terminate... */
-	exit(0);
-
-	/* so the compiler doesn't complain.. */
-	return;
-        }
-
-
-
 /* free all allocated memory */
 void free_memory(void){
 	command *this_command;
@@ -1031,18 +1206,21 @@ void free_memory(void){
 	this_command=command_list;
 	while(this_command!=NULL){
 		next_command=this_command->next;
-		free(this_command->command_name);
-		free(this_command->command_line);
+		if(this_command->command_name)
+			free(this_command->command_name);
+		if(this_command->command_line)
+			free(this_command->command_line);
 		free(this_command);
 		this_command=next_command;
 	        }
 
+	command_list=NULL;
+
 	return;
         }
 
 
 
-
 /* executes a system command via popen(), but protects against timeouts */
 int my_system(char *command,int timeout,int *early_timeout,char *output,int output_length){
         pid_t pid;
@@ -1270,10 +1448,10 @@ int drop_privileges(char *user, char *group){
 		else
 			uid=(uid_t)atoi(user);
 			
-#ifdef HAVE_INITGROUPS
-
+		/* set effective user ID if other than current EUID */
 		if(uid!=geteuid()){
 
+#ifdef HAVE_INITGROUPS
 			/* initialize supplementary groups */
 			if(initgroups(user,gid)==-1){
 				if(errno==EPERM)
@@ -1283,11 +1461,11 @@ int drop_privileges(char *user, char *group){
 					return ERROR;
 			                }
 	                        }
-		        }
 #endif
 
-		if(setuid(uid)==-1)
-			syslog(LOG_ERR,"Warning: Could not set effective UID=%d",(int)uid);
+			if(setuid(uid)==-1)
+				syslog(LOG_ERR,"Warning: Could not set effective UID=%d",(int)uid);
+		        }
 	        }
 
 	return OK;
@@ -1323,7 +1501,7 @@ int write_pid_file(void){
 
 			/* previous process is still running */
 			else{
-				syslog(LOG_ERR,"There's already a nrpe server running.");
+				syslog(LOG_ERR,"There's already an NRPE server running (PID %lu).  Bailing out...",(unsigned long)pid);
 				return ERROR;
 			        }
 		        }
@@ -1334,9 +1512,32 @@ int write_pid_file(void){
 		sprintf(pbuf,"%d\n",(int)getpid());
 		write(fd,pbuf,strlen(pbuf));
 		close(fd);
+		wrote_pid_file=TRUE;
 	        }
 	else{
-		syslog(LOG_ERR,"Cannot write to pidfile '%s'.",pid_file);
+		syslog(LOG_ERR,"Cannot write to pidfile '%s' - check your privileges.",pid_file);
+	        }
+
+	return OK;
+        }
+
+
+
+/* remove pid file */
+int remove_pid_file(void){
+
+	/* no pid file was specified */
+	if(pid_file==NULL)
+		return OK;
+
+	/* pid file was not written */
+	if(wrote_pid_file==FALSE)
+		return OK;
+
+	/* remove existing pid file */
+	if(unlink(pid_file)==-1){
+		syslog(LOG_ERR,"Cannot remove pidfile '%s' - check your privileges.",pid_file);
+		return ERROR;
 	        }
 
 	return OK;
@@ -1362,6 +1563,59 @@ int check_privileges(void){
 
 
 
+/* handle signals (parent process) */
+void sighandler(int sig){
+	static char *sigs[]={"EXIT","HUP","INT","QUIT","ILL","TRAP","ABRT","BUS","FPE","KILL","USR1","SEGV","USR2","PIPE","ALRM","TERM","STKFLT","CHLD","CONT","STOP","TSTP","TTIN","TTOU","URG","XCPU","XFSZ","VTALRM","PROF","WINCH","IO","PWR","UNUSED","ZERR","DEBUG",(char *)NULL};
+	int i;
+	char temp_buffer[MAX_INPUT_BUFFER];
+
+	if(sig<0)
+		sig=-sig;
+
+	for(i=0;sigs[i]!=(char *)NULL;i++);
+
+	sig%=i;
+
+	/* we received a SIGHUP, so restart... */
+	if(sig==SIGHUP){
+
+		sigrestart=TRUE;
+
+		syslog(LOG_NOTICE,"Caught SIGHUP - restarting...\n");
+	        }
+
+	/* else begin shutting down... */
+	if(sig==SIGTERM){
+
+		/* if shutdown is already true, we're in a signal trap loop! */
+		if(sigshutdown==TRUE)
+			exit(STATE_CRITICAL);
+
+		sigshutdown=TRUE;
+
+		syslog(LOG_NOTICE,"Caught SIG%s - shutting down...\n",sigs[sig]);
+	        }
+
+	return;
+        }
+
+
+
+/* handle signals (child processes) */
+void child_sighandler(int sig){
+
+	/* free all memory we allocated */
+	free_memory();
+
+	/* terminate */
+	exit(0);
+	
+	/* so the compiler doesn't complain... */
+	return;
+        }
+
+
+
 /* tests whether or not a client request is valid */
 int validate_request(packet *pkt){
         u_int32_t packet_crc32;
